"""This is a one file library with main function initialize_project_logger
which can be used as first logger set up for new py packages
Manual can be found here: https://github.com/stas-prokopiev/basic_package_logger
"""
import sys
import os
import io
import logging

# Define constants with msg formats
# File msg formats
STR_DEBUG_FILE_FORMAT = (
    "{"
    "'levelname':'%(levelname)s', "
    "'asctime':'%(asctime)s', "
    "'filename':'%(filename)s', "
    "'name':'%(name)s', "
    "'funcName':'%(funcName)s', "
    "'lineno':'%(lineno)d', "
    "'message':'%(message)s'"
    "}"
)

STR_ERROR_FILE_FORMAT = STR_DEBUG_FILE_FORMAT
# stdout msg formats
STR_DEBUG_FORMAT = '[%(levelname)s] %(message)s'
STR_INFO_FORMAT = '%(message)s'
STR_WARNING_FORMAT = '[%(levelname)s] %(message)s'
STR_ERROR_FORMAT = (
    '[%(levelname)s: %(asctime)s:%(filename)s:'
    '%(name)s:%(funcName)s:%(lineno)d] %(message)s')


class OnlyLowerLevelFilter():
    """Define filter to show only logs with level lower than given level
    """
    def __init__(self, level):
        self.level = level

    def filter(self, record):
        """Filter messages with level lower than given"""
        return record.levelno < self.level


def initialize_project_logger(
        name,
        path_dir_where_to_store_logs="",
        is_stdout_debug=False,
        is_to_propagate_to_root_logger=False,
):
    """function returns a perfectly set up logger for the new package"""
    logger_obj = logging.getLogger(name)
    # If logger already exists then just return it
    if logger_obj.handlers:
        return logger_obj
    # Create and set basic settings for logger
    logger_obj.setLevel(20-int(is_stdout_debug)*10)
    logger_obj.propagate = is_to_propagate_to_root_logger
    # If root logger has not been set
    # Or user don't know what is logging at all
    # logging.basicConfig(file="/dev/null", level=0)  # Linux
    # logging.basicConfig(file="NUL", level=0)  # Win
    logging.basicConfig(stream=io.StringIO(), level=0)  # Any but eats memory
    #####
    # 1) Set up stdout logs
    # 1.0) Add debug handler if necessary
    if is_stdout_debug:
        debug_handler = logging.StreamHandler(sys.stdout)
        debug_handler.setLevel(level=0)
        debug_handler.setFormatter(logging.Formatter(STR_DEBUG_FORMAT))
        debug_handler.addFilter(OnlyLowerLevelFilter(20))
        logger_obj.addHandler(debug_handler)
    # 1.1) Add info handler
    info_handler = logging.StreamHandler(sys.stdout)
    info_handler.setLevel(level=20)
    info_handler.setFormatter(logging.Formatter(STR_INFO_FORMAT))
    info_handler.addFilter(OnlyLowerLevelFilter(30))
    logger_obj.addHandler(info_handler)
    # 1.2) Add warning handler
    warning_handler = logging.StreamHandler(sys.stdout)
    warning_handler.setLevel(level=30)
    warning_handler.setFormatter(logging.Formatter(STR_WARNING_FORMAT))
    warning_handler.addFilter(OnlyLowerLevelFilter(40))
    logger_obj.addHandler(warning_handler)
    # 1.3) Add error handler
    error_handler = logging.StreamHandler(sys.stderr)
    error_handler.setLevel(level=40)
    error_handler.setFormatter(logging.Formatter(STR_ERROR_FORMAT))
    logger_obj.addHandler(error_handler)
    if not path_dir_where_to_store_logs:
        return None
    if not os.path.isdir(path_dir_where_to_store_logs):
        raise TypeError(
            "Wrong type of argument 'path_dir_where_to_store_logs'\n"
            "Expected to get string with path to some directory "
            "But got: " + str(path_dir_where_to_store_logs)
        )
    #####
    # 2) Set up file handlers for logger
    # Create folder for Logs
    str_path_to_logs_dir = os.path.join(path_dir_where_to_store_logs, "Logs")
    if not os.path.isdir(str_path_to_logs_dir):
        os.makedirs(str_path_to_logs_dir)
    # 2.1) Debug handler
    debug_file_handler = logging.handlers.RotatingFileHandler(
        os.path.join(str_path_to_logs_dir, "debug.log"),
        maxBytes=10000,
        backupCount=1
    )
    debug_file_handler.setLevel(level=0)
    debug_file_handler.setFormatter(logging.Formatter(STR_DEBUG_FILE_FORMAT))
    logger_obj.addHandler(debug_file_handler)
    # 2.1) Warnings and above handler
    warnings_file_handler = logging.handlers.RotatingFileHandler(
        os.path.join(str_path_to_logs_dir, "errors.log"),
        maxBytes=10000,
        backupCount=1
    )
    warnings_file_handler.setLevel(level=30)
    warnings_file_handler.setFormatter(
        logging.Formatter(STR_ERROR_FILE_FORMAT))
    logger_obj.addHandler(warnings_file_handler)



"""Module with functions to create execution prices"""
# Standard library imports
import logging
from typing import Optional, Any, Union
import datetime

# Third party imports
import pandas as pd

# Local imports

# Global constants
LOGGER = logging.getLogger(__name__)


def create_df_simple_execution_prices(
        df_prices_full : pd.DataFrame,
        data_frequency="30min",  # 30min, 1h, 3h, 6h, 1d
        offset="-3min",  # "-3min"
        td_trading_delay : Optional[datetime.timedelta]=None,
        td_execution_duration : Optional[datetime.timedelta]=None,
):
    """Backtest positions to understand if they can generate PNL

    Args:
        df_positions_short (pd.DataFrame): Positions we want to take
        df_prices_full (pd.DataFrame): Prices of assets in higher resolution
        td_trading_delay (datetime.timedelta): \
            Delay needed to get into the wanted positions
        td_execution_duration ([datetime.timedelta]): \
            How long should the execution take

    Returns:
        pd.DataFrame: columns with different PNLs generated at every tick
    """
    LOGGER.info("Create simple execution prices")
    #####
    # Shift prices on asked delay
    LOGGER.info("---> Shift prices on delay in the past: %s", td_trading_delay)
    # print(df_prices_full.tail(5)["AAVEUSDT"])
    LOGGER.info(
        "------> First datetime in index BEFORE: %s",
        df_prices_full.index.to_list()[0]
    )
    df_prices_full = df_prices_full.shift(freq=-td_trading_delay)
    # print(df_prices_full.tail(5)["AAVEUSDT"])
    LOGGER.info(
        "------> First datetime in index AFTER: %s",
        df_prices_full.index.to_list()[0]
    )
    LOGGER.info("------> Done")
    #####
    # Calculate rolling mean price over the given duration
    LOGGER.info(
        "---> Calculate mean price over the execution duration: %s",
        td_execution_duration)
    ## Reverse the df with prices because rolling mean price
    ## Should go into future not the past and then at the end reverse back
    # print(df_prices_full.tail(7)["AAVEUSDT"])
    df_real_exec_price_full = df_prices_full[::-1].rolling(
        td_execution_duration, min_periods=2).mean()[::-1]
    # print(df_real_exec_price_full.tail(7)["AAVEUSDT"])
    LOGGER.info("------> Done")
    #####
    # Resample to the asked frequency
    LOGGER.info(
        "---> Resample to: frequency=%s offset=%s", data_frequency, offset)
    # print(df_real_exec_price_full.tail(33)["AAVEUSDT"])
    df_real_exec_prices_short = resample_df_to_asked_frequency(
        df_real_exec_price_full,
        data_frequency=data_frequency,
        offset=offset,
    )
    LOGGER.info("------> Done")
    return df_real_exec_prices_short


def resample_df_to_asked_frequency(
        df : pd.DataFrame,
        data_frequency="30min",  # 30min, 1h, 3h, 6h, 1d
        offset="-3min",  # "-3min"
):
    """"""
    df_resampled = df.resample(
        data_frequency,
        label="right",  # left, right
        closed="right",  # left, right
        offset=offset
    )
    return df_resampled.last()




def get_df_perfect_execution_prices(
        df_positions_short : pd.DataFrame,
        df_prices_full : pd.DataFrame,
        td_trading_delay : Optional[datetime.timedelta]=None,
):
    """TODO"""
    pass


def _convert_full_prices_df_to_short(
        df_positions_short : pd.DataFrame,
        df_prices_full : pd.DataFrame,
) -> pd.DataFrame:
    """Convert prices in high resolution to the resolution of short DFs"""
    set_index_short = set(df_positions_short.index.tolist())
    set_index_full = set(df_prices_full.index.tolist())
    set_missing_ticks_in_full = set_index_short - set_index_full
    set_common_ticks = set_index_short.intersection(set_index_full)

    if set_missing_ticks_in_full:
        LOGGER.warning("Missing ticks: %d", len(set_missing_ticks_in_full))

    if len(set_missing_ticks_in_full) > 5:
        list_missing_ticks = sorted(set_missing_ticks_in_full)
        LOGGER.warning("Missing ticks: %s", list_missing_ticks[:20])
        raise ValueError(
            f"Unable to convert full df into short format as "
            f"{len(set_missing_ticks_in_full)}/{len(set_index_short)} "
            "ticks missing"
        )
    return df_prices_full.loc[sorted(set_common_ticks)]



""""""
from . import logger

logger.initialize_project_logger(
    name=__name__,
    path_dir_where_to_store_logs="",
    is_stdout_debug=True,
    is_to_propagate_to_root_logger=False,
)
# Standard library imports

# Third party imports

# Local imports
from .backtester import run_backtest
from .simple_execution import create_df_simple_execution_prices

# Global constants

__all__ = ["run_backtest", "create_df_simple_execution_prices"]








# def calc_execution_price_rough(
#         # df_positions_short : pd.DataFrame,
#         df_execution_prices_full : pd.DataFrame,
#         td_execution_duration : datetime.timedelta
# ) -> pd.DataFrame:
#     """Get execution price as mean price over execution duration

#     Args:
#         df_positions_short (pd.DataFrame): Positions we want to take
#         df_execution_prices_full (pd.DataFrame): Prices of assets in higher resolution
#         td_execution_duration ([datetime.timedelta]): \
#             How long should the execution take

#     Returns:
#         pd.DataFrame: Prices by which asset can be bought at any moment
#     """
#     # Reverse the df with prices because rolling mean price
#     # Should go into future not the past and then at the end reverse back
#     LOGGER.debug("------> Calculate full execution prices")

#     if td_execution_duration:
#         df_exec_price_full = df_execution_prices_full[::-1].rolling(
#             td_execution_duration, min_periods=2).mean()[::-1]
#     else:
#         df_exec_price_full = df_execution_prices_full[::-1].rolling(2).mean()[::-1]


#     return df_exec_price_full
#     # LOGGER.debug("------> Convert them into short format")
#     # # Convert Execution prices to the short format
#     # df_exec_prices_short = _convert_full_prices_df_to_short(
#     #     df_positions_short, df_exec_price_full)
#     # LOGGER.debug("---------> Done")
#     # return df_exec_prices_short




# def _convert_full_prices_df_to_short(
#         df_positions_short,
#         df_execution_prices_full
# ):
#     """Convert prices in high resolution to the resolution of short DFs"""
#     from tqdm.auto import tqdm
#     start = time.time()
#     set_index_short = set(df_positions_short.index.tolist())
#     set_index_full = set(df_execution_prices_full.index.tolist())
#     new_full_index_rev_sorted = reversed(
#         sorted(set_index_short.union(set_index_full)))

#     print(0, time.time() - start)

#     dict_df_execution_prices_short = {}

#     dict_df = df_execution_prices_full.to_dict(orient='list')
#     print(1, time.time() - start)
#     # dict_df = df_execution_prices_full.to_dict(orient='dict')
#     print(1.5, time.time() - start)

#     for ticker in tqdm(dict_df):
#         list_values_for_ticker = reversed(dict_df[ticker])
#         dict_value_by_timestamp_short = {}
#         last_value = None
#         index = 0
#         for timestamp in new_full_index_rev_sorted:
#             if timestamp in set_index_full:
#                 last_value = list_values_for_ticker[index]
#                 index += 1
#             if timestamp in set_index_short:
#                 dict_value_by_timestamp_short[timestamp] = last_value
#         dict_df_execution_prices_short[ticker] = dict_value_by_timestamp_short
#     print(2, time.time() - start)
#     df = pd.DataFrame.from_dict(dict_df_execution_prices_short)
#     print(3, time.time() - start)
#     return df


# def _convert_full_prices_df_to_short(
#         df_positions_short,
#         df_execution_prices_full
# ):
#     """
#     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#     DO NOT DELETE, this is a benchmark that my realization works correctly
#     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#     Convert prices in high resolution to the resolution of short DFs
#     """
#     df_positions_empty = pd.DataFrame(
#         index=df_positions_short.index, columns=df_positions_short.columns)
#     df_prices_full_size = pd.concat(
#         [df_execution_prices_full, df_positions_empty], axis=1)
#     df_prices_full_size.sort_index(inplace=True)



#     df_grouped = df_prices_full_size.groupby(by=df_prices_full_size.index.date)
#     for name, group in tqdm(df_grouped):
#         group.bfill(inplace=True)


#     # df_prices_full_size.bfill(inplace=True)
#     df_prices_short = df_prices_full_size.loc[df_positions_short.index]
#     return df_prices_short






# def _backfill_values_in_dataframe(df):
#     """"""
#     dict_df = df.to_dict(orient='list')
#     for ticker in dict_df:
#         print(ticker)
#         break

#     return pd.DataFrame.from_dict(dict_df)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['positions_backtester']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.4.0,<2.0.0', 'tqdm>=4.64.0,<5.0.0']

setup_kwargs = {
    'name': 'positions-backtester',
    'version': '0.1.5',
    'description': 'Package to backtest trading strategies (dataframe with positions) with execution costs modeling',
    'long_description': '=====================\npositions_backtester\n=====================\n\n.. image:: https://img.shields.io/github/last-commit/stas-prokopiev/positions_backtester\n   :target: https://img.shields.io/github/last-commit/stas-prokopiev/positions_backtester\n   :alt: GitHub last commit\n\n.. image:: https://img.shields.io/github/license/stas-prokopiev/positions_backtester\n    :target: https://github.com/stas-prokopiev/positions_backtester/blob/master/LICENSE.txt\n    :alt: GitHub license<space><space>\n\n.. image:: https://img.shields.io/pypi/v/positions_backtester\n   :target: https://img.shields.io/pypi/v/positions_backtester\n   :alt: PyPI\n\n.. image:: https://img.shields.io/pypi/pyversions/positions_backtester\n   :target: https://img.shields.io/pypi/pyversions/positions_backtester\n   :alt: PyPI - Python Version\n\n\n.. contents:: **Table of Contents**\n\nShort Overview.\n=========================\npositions_backtester is a python package (**py>=3.7**) to backtest trading strategies (dataframe with positions) with execution costs modeling\n\n| This package is trying to solve a problem of slow trading with fast data.\n| Let\'s say that you want to update your trading position once in a hour, day, week, ...\n| But you have data with much higher time resolution - minutes, seconds, miliseconds\n| Then you can give the wanted positions dataframe with the tick with which you want to trade\n| And higher resolution will be used to calculate approximate execution prices\n| (Which are just the mean price over the choosen execution time period)\n\nInstallation via pip:\n======================\n\n.. code-block:: bash\n\n    pip install positions_backtester\n\n\nHow to backtest your dataframe with positions\n================================================\n\n.. code-block:: python\n\n    from positions_backtester import run_backtest\n\n    df_backtest_res = run_backtest(\n        df_positions,\n        df_real_exec_prices,\n        is_to_neutralize=True,\n        const_trading_fees_percent=0.01,\n    )\n\n!!! **df_real_exec_prices should have the same indices as df_positions**\n\nArguments:\n\n#. **df_positions**:\n    | pd.DataFrame\n    | Positions we want to take with the frequency with which we want to change our positions\n#. **df_real_exec_prices**:\n    | pd.DataFrame\n    | Prices by which we can buy assets at the given moment\n#. **is_to_neutralize=True,**:\n    | bool\n    | Flag if to have long-short equal positions\n#. **td_trading_delay=None**:\n    | datetime.timedelta\n    | Delay needed to calculate the wanted positions\n#. **td_execution_duration**:\n    | datetime.timedelta\n    | How long should the execution take\n    | Execution price will be the mean price over execution time period\n#. **const_trading_fees_percent**:\n    | float\n    | Broker trading fees\n\n\nInput example: **df_positions_short**\n-----------------------------------------------\n\n..\n    This is a comment\n\n    .. raw:: html\n\n    <embed>\n        <table border="1" class="dataframe">\n        <thead>\n            <tr style="text-align: right;">\n            <th></th>\n            <th>asset_1</th>\n            <th>asset_2</th>\n            <th>asset_3</th>\n            </tr>\n            <tr>\n            <th>Close datetime</th>\n            <th></th>\n            <th></th>\n            <th></th>\n            </tr>\n        </thead>\n        <tbody>\n            <tr>\n            <th>2021-07-06 22:00:00+00:00</th>\n            <td>0.285602</td>\n            <td>NaN</td>\n            <td>NaN</td>\n            </tr>\n            <tr>\n            <th>2021-07-06 23:00:00+00:00</th>\n            <td>0.296204</td>\n            <td>NaN</td>\n            <td>NaN</td>\n            </tr>\n            <tr>\n            <th>2021-07-07 00:00:00+00:00</th>\n            <td>0.294426</td>\n            <td>NaN</td>\n            <td>NaN</td>\n            </tr>\n        </tbody>\n        </table>\n    </embed>\n\n\n\nOutput: df_backtest_res\n-----------------------------------------------\n\n..\n    This is a comment\n    .. raw:: html\n\n    <embed>\n        <table border="1" class="dataframe">\n        <thead>\n            <tr style="text-align: right;">\n            <th></th>\n            <th>PNL before_costs</th>\n            <th>PNL execution_fee</th>\n            <th>PNL const_trading_fee</th>\n            <th>PNL after_costs</th>\n            <th>PNL half_costs</th>\n            <th>trading_volume</th>\n            <th>booksize</th>\n            <th>max weight</th>\n            <th>long count</th>\n            <th>short count</th>\n            <th>long value</th>\n            <th>short value</th>\n            </tr>\n            <tr>\n            <th>Close datetime</th>\n            <th></th>\n            <th></th>\n            <th></th>\n            <th></th>\n            <th></th>\n            <th></th>\n            </tr>\n        </thead>\n        <tbody>\n            <tr>\n            <th>2021-07-06 20:00:00+00:00</th>\n            <td>-0.002108</td>\n            <td>-0.004361</td>\n            <td>0.000003</td>\n            <td>0.002250</td>\n            <td>0.000071</td>\n            <td>0.034720</td>\n            <td>1.0</td>\n            <td>0.14</td>\n            <td>33</td>\n            <td>200</td>\n            <td>0.5</td>\n            <td>0.5</td>\n            </tr>\n            <tr>\n            <th>2021-07-06 21:00:00+00:00</th>\n            <td>-0.005282</td>\n            <td>-0.000222</td>\n            <td>0.000005</td>\n            <td>-0.005065</td>\n            <td>-0.005174</td>\n            <td>0.053568</td>\n            <td>1.0</td>\n            <td>0.11</td>\n            <td>32</td>\n            <td>198</td>\n            <td>0.5</td>\n            <td>0.5</td>\n            </tr>\n        </tbody>\n        </table>\n    </embed>\n\n\n\n\nHow to create execution prices\n===================================\n\n.. code-block:: python\n\n    from positions_backtester import create_df_simple_execution_prices\n\n    df_real_exec_prices = create_df_simple_execution_prices(\n        df_prices_full,\n        data_frequency="30min",  # 30min, 1h, 3h, 6h, 1d\n        offset="-3min",  # "-3min"\n        td_trading_delay=datetime.timedelta(minutes=3),\n        td_execution_duration=datetime.timedelta(minutes=2),\n    )\n\n\nArguments:\n\n#. **df_prices_full**:\n    | pd.DataFrame\n    | Prices of assets in higher resolution\n    | Higher resolution needed for more pricise execution evaluation\n#. **td_trading_delay=None**:\n    | datetime.timedelta\n    | Delay (Time) needed to calculate the wanted positions\n#. **td_execution_duration**:\n    | datetime.timedelta\n    | How long should the execution take\n    | Execution price will be the mean price over execution time period\n\n\nInputs:\n-----------------------------------------------\n\n\ndf_prices_full\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n\n..\n    This is a comment\n    .. raw:: html\n\n    <embed>\n        <table border="1" class="dataframe">\n        <thead>\n            <tr style="text-align: right;">\n            <th></th>\n            <th>asset_1</th>\n            <th>asset_2</th>\n            <th>asset_3</th>\n            </tr>\n            <tr>\n            <th>Close datetime</th>\n            <th></th>\n            <th></th>\n            <th></th>\n            </tr>\n        </thead>\n        <tbody>\n            <tr>\n            <th>2021-07-06 23:57:00+00:00</th>\n            <td>317.86</td>\n            <td>57.00</td>\n            <td>15.488</td>\n            </tr>\n            <tr>\n            <th>2021-07-06 23:58:00+00:00</th>\n            <td>317.11</td>\n            <td>57.04</td>\n            <td>15.480</td>\n            </tr>\n            <tr>\n            <th>2021-07-06 23:59:00+00:00</th>\n            <td>316.49</td>\n            <td>57.01</td>\n            <td>15.459</td>\n            </tr>\n        </tbody>\n        </table>\n    </embed>\n\n\n\nFormulas\n===========================\n\n| PNL before_costs = (previous_position) * (price_change_%)\n| trading_volume = abs(new_wanted_position - previous_position)\n| PNL const_trading_fee = trading_volume * broker_commision\n| PNL after_costs = PNL before_costs - PNL const_trading_fee\n| PNL half_costs = PNL before_costs - PNL const_trading_fee/2.0\n\nLinks\n=====\n\n    * `PYPI <https://pypi.org/project/positions_backtester/>`_\n    * `GitHub <https://github.com/stas-prokopiev/positions_backtester>`_\n\nProject local Links\n===================\n\n    * `CHANGELOG <https://github.com/stas-prokopiev/positions_backtester/blob/master/CHANGELOG.rst>`_.\n    * `CONTRIBUTING <https://github.com/stas-prokopiev/positions_backtester/blob/master/CONTRIBUTING.rst>`_.\n\nContacts\n========\n\n    * Email: stas.prokopiev@gmail.com\n    * `vk.com <https://vk.com/stas.prokopyev>`_\n    * `Facebook <https://www.facebook.com/profile.php?id=100009380530321>`_\n\nLicense\n=======\n\nThis project is licensed under the MIT License.',
    'author': 'stanislav',
    'author_email': 'stas.prokopiev@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/stas-prokopiev/positions_backtester',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Main module of the package"""
# Standard library imports
import logging
from typing import Optional, Any, Union
import datetime
import time

# Third party imports
import pandas as pd
from tqdm.auto import tqdm

# Local imports

# Global constants
LOGGER = logging.getLogger(__name__)


def run_backtest(
        df_positions : pd.DataFrame,
        df_real_exec_prices : pd.DataFrame,
        df_perfect_exec_prices : pd.DataFrame = None,
        is_to_neutralize : bool=True,
        const_trading_fees_percent : float=0.01,
) -> pd.DataFrame:
    """Backtest positions to understand if they can generate PNL

    Args:
        df_positions (pd.DataFrame): Positions we want to take
        df_execution_prices_full (pd.DataFrame): Prices of assets in higher resolution
        is_to_neutralize (bool): Flag if to have long-short equal positions
        const_trading_fees_percent (float): Broker commission fee in percent

    Returns:
        pd.DataFrame: columns with different PNLs generated at every tick
    """
    LOGGER.debug("Run backtest")
    # print(df_positions.T.tail(3).T.tail(3).to_html())
    df_backtest_res = pd.DataFrame()
    # Neutralize
    if is_to_neutralize:
        LOGGER.debug("---> Neutralize positions for every datetime")
        df_positions = _neutralize(df_positions)
    # Scale to 1.0
    LOGGER.debug("---> Scale positions to 1 for every datetime")
    df_positions = _scale(df_positions)
    # Calculate holding pnl
    # which is a pnl generated by positions taken in the past
    LOGGER.debug("---> Calculate holding PNL")
    df_backtest_res["PNL before_costs"] = _calc_ser_holding_pnl(
        df_positions, df_real_exec_prices)
    # DataFrame how we would like to change our current positions
    LOGGER.debug("---> Calculate the wanted position change on every tick")
    df_pos_change_wanted = df_positions - df_positions.shift(1)
    if df_perfect_exec_prices is not None:
        df_lost_value = (
            (df_real_exec_prices - df_perfect_exec_prices) /
            df_perfect_exec_prices
        )
        df_execution_fee = df_pos_change_wanted * df_lost_value
        df_backtest_res["PNL execution_fee"] = df_execution_fee.sum(axis=1)
    # Volume traded at the current tick
    df_backtest_res["trading_volume"] = \
        df_pos_change_wanted.abs().sum(axis=1)
    # Brokers trading commission
    df_backtest_res["PNL const_trading_fee"] = (
        df_backtest_res["trading_volume"] *
        const_trading_fees_percent / 100.0
    )
    # Add column with with final PNL results
    df_backtest_res["PNL after_costs"] = (
        df_backtest_res["PNL before_costs"] -
        # df_backtest_res["PNL execution_fee"] -
        df_backtest_res["PNL const_trading_fee"]
    )
    df_backtest_res["PNL half_costs"] = (
        df_backtest_res["PNL before_costs"] +
        df_backtest_res["PNL after_costs"]
    ) / 2.0
    # Add more columns with info
    df_backtest_res["booksize"] = df_positions.abs().sum(axis=1)
    df_backtest_res["max weight"] = df_positions.abs().max(axis=1)
    df_backtest_res["long count"] = \
        df_positions[df_positions > 0].count(axis=1)
    df_backtest_res["short count"] = \
        df_positions[df_positions < 0].count(axis=1)
    df_backtest_res["long value"] = \
        df_positions[df_positions > 0].sum(axis=1).abs()
    df_backtest_res["short value"] = \
        df_positions[df_positions < 0].sum(axis=1).abs()
    return change_final_columns_order(df_backtest_res)


def change_final_columns_order(df_backtest_res):
    """"""
    columns_ordered = []
    columns_pnl = []
    for column in df_backtest_res.columns:
        if "PNL" in column:
            columns_pnl.append(column)
            continue
        columns_ordered.append(column)
    df_backtest_res = df_backtest_res[columns_pnl + columns_ordered]
    return df_backtest_res


def _calc_ser_holding_pnl(
        df_positions : pd.DataFrame,
        df_execution_prices : pd.DataFrame
) -> pd.Series:
    """Calculate PNL generated by holding positions taken at last tick"""
    LOGGER.debug("------> Shift positions on 1 tick backward")

    df_previous_pos = df_positions.shift(1)
    df_prices_change_pct = \
        (df_execution_prices / df_execution_prices.shift(1)) - 1.0
    LOGGER.debug(
        "------> Multiply previous tick positions on price change percent")
    df_holding_pnl = df_previous_pos.multiply(df_prices_change_pct)
    return df_holding_pnl.sum(axis=1)


def _neutralize(df_positions : pd.DataFrame) -> pd.DataFrame:
    """Neutralize positions to be long-short equal"""
    return df_positions.sub(df_positions.mean(axis=1), axis=0)


def _scale(df_positions : pd.DataFrame) -> pd.DataFrame:
    """
    Scale to have sum of absolute positions at every tick equals to 1.0
    """
    return df_positions.div(df_positions.abs().sum(axis=1), axis=0)
